"""
Some functions to support with plotting in Python based on the *matplotlib* module
"""

import itertools
import logging
import re

_logger = logging.getLogger()


# definition of colors
c_sim = dict(
    blue="#0083D1",
    lightblue="#69ABD2",
    red="#DD2E1A",
    lightred="#DD877E",
    green="#00CC00",
    darkcyan="#2eb8b8",
)


def analyse_annotations(annotation):
    """Analyze the string `annotation` which compactly sets the properties of a label
    string such as position, size, and color.

    Parameters
    ----------
    annotation : str
        label[@(xp, yp)[s10][a0][cRRGGBB]]

    Returns
    -------
    tuple of strings and integers
        text (str), x position (float) y position (float), color (str), size (int),
        axis (int)

    Notes
    -----
    * The annotation string can be just given as a label. This label optionally can be
      extended with a '@' sign to include more information, like the location (xp,
     yp), the size with s10, the color with c and the axis with a and a integer

    * The annotation format is mostly used as a compact way to provide information on
      the annotation via a user input file

    * When specifying a color with the c construct, make sure that you put the color at
      the end of the *annotation* string.

    * In the Parameter scription we used a hex formulation for the color like cFFAA00.
      However, also the color names like cblue or cred are allowed

    * In case you add chmc:red, the hmc definition of red will be used. Other hmc color
      definitions are:

         - blue
         - lightblue
         - red
         - lightred
         - green
         - darkcyan

    * Also we can add the xkcd_ color definitions as described in the matplotlib_ manual

    .. _xkcd:
        https://xkcd.com/color/rgb/

    .. _matplotlib:
        https://matplotlib.org/users/colors.html

    Examples
    --------

    Simple example of a label with all the properties set to default

    >>> analyse_annotations("simple_label")
    ('simple_label', 0.0, 0.0, 'black', 12, 0)

    place a label at position x=0.1, y=0.4

    >>> analyse_annotations("more_complex@0.1,0.4")
    ('more_complex', 0.1, 0.4, 'black', 12, 0)

    place a label at position x=0.8, y=0.9 with color red. Note that we need to add
    brackets around the location

    >>> analyse_annotations("colored_label@(0.8,0.9)cred")
    ('colored_label', 0.8, 0.9, 'red', 12, 0)

    place a label at position x=0.8, y=0.9 with color red. Note that we need to add
    brackets around the location

    >>> analyse_annotations("small_label@(0.8,0.9)s8")
    ('small_label', 0.8, 0.9, 'black', 8, 0)

    Place a label at position x=0.8, y=0.9 with color red and size 16. This time we need
    to add the color label at the end to extract it correcly

    >>> analyse_annotations("large_colored_label_in_axis_2@(0.8,0.9)s16a2cred")
    ('large_colored_label_in_axis_2', 0.8, 0.9, 'red', 16, 2)

    Finally lets show how you use more colours

    >>> analyse_annotations("label@(0.8,0.9)s16a2chmc:red")
    ('label', 0.8, 0.9, '#DD2E1A', 16, 2)

    As you can see, the hex code is returned.

    To set the xkcd colors do

    >>> analyse_annotations("label@(0.8,0.9)s16a2cxkcd:red")
    ('label', 0.8, 0.9, 'xkcd:red', 16, 2)

    This color value 'xkcd:red' is understood by all matplotlib routines
    """
    lx = 0.0
    ly = 0.0
    color = "black"
    axis = 0
    size = 12
    pos = f"({lx},{ly})"
    # first replace all white spaces from the string, as it is not allowed
    try:
        # see if there is an @ sign indicating that the position is specified
        text, rest = annotation.split("@")
        try:
            # after the @ sign we start with the position between brackets (,). Find it
            pos, rest = re.sub("[(|)]", " ", rest).split()

            # now the rest is only a size s18 and a color cred or c#FFAA00 (hexa code).
            # First get the size, then the color
            size_pattern = r"s(\d+)"
            axis_pattern = r"a(\d)"
            color_pattern = "c(.*)"
            m = re.search(size_pattern, rest)
            if bool(m):
                size = int(m.group(1))
                rest = re.sub(size_pattern, "", rest)
            m = re.match(axis_pattern, rest)
            if bool(m):
                try:
                    axis = int(m.group(1))
                except ValueError:
                    _logger.warning("axis must by integers. Set zero")
                rest = re.sub(axis_pattern, "", rest)
            m = re.match(color_pattern, rest)
            if bool(m):
                color = m.group(1)

        except ValueError:
            # in case of a value error we did not have a rest, so try again without
            # split
            pos = re.sub("[(|)]", " ", rest)
        finally:
            lx, ly = pos.split(",")
            lx = float(lx)
            ly = float(ly)
    except ValueError:
        # there as no @ sign: just return the text value with all the rest the defaults
        text = annotation

    return text, lx, ly, color, size, axis


def clean_up_artists(axis, artist_list):
    """Remove all the artists stored in the `artist_list` belonging to the `axis`.


    Parameters
    ----------
    axis : :class:matplotlib.axes.Axes
        Clean Artists (ie. items added to a matplotlib plot) belonging to this axis
    artist_list : list
        List of artist to remove.

    Notes
    -----

    In case of animation of complex plots with contours and labels (such as a timer) we
    sometimes need to take care of removing all the Artists which are changing every
    time step.
    This function takes care of that. It also also ensured that we are not running out
    of memory when too many Artists are added

    Examples
    --------

    >>> from matplotlib.pyplot import subplots
    >>> from numpy.random import random_sample

    Create a list which we use to store all the artists which need to be cleaned

    >>> artist_list = list()

    Create a plot of some random data

    >>> fig, ax = subplots(ncols=1, nrows=1)
    >>> data = random_sample((20, 30))
    >>> cs = ax.contourf(data)

    Store the contour Artist in a list

    >>> artist_list.append(cs)

    Now clean it again

    >>> clean_up_artists(ax, artist_list)

    """

    for artist in artist_list:
        try:
            # fist attempt: try to remove collection of contours for instance
            while artist.collections:
                for col in artist.collections:
                    artist.collections.remove(col)
                    try:
                        axis.collections.remove(col)
                    except ValueError:
                        pass

                artist.collections = []
                axis.collections = []
        except AttributeError:
            pass

        # second attempt, try to remove the text
        try:
            artist.remove()
        except (AttributeError, ValueError):
            pass


def clean_up_plot(artist_list):
    """A small script to clean up all lines or other items of a matplot lib plot.


    Parameters
    ----------
    artist_list :
        a list of items to clean up

    Notes
    -----

    Necessary if you want to loop over multiple plot and maintain the axes and only
    update the data.
    Basically this does the same as `clean_up_artists`

    """
    n_cleaned = 0
    for ln in artist_list:
        n_cleaned += 1
        try:
            ln.pop(0).remove()
        except (IndexError, AttributeError):
            try:
                ln.remove()
            except (ValueError, TypeError, AttributeError):
                del ln
        else:
            n_cleaned -= 1
            _logger.debug("All clean up failed. ")
    artist_list = []
    return artist_list


def sub_plot_axis_to_2d(axis, n_rows=1, n_cols=1):
    """
    Turn the axis return value of matplotlibs subfigures in a 2D list, regardless of the
    number of rows and columns

    Parameters
    ----------
    axis : list or :class:matplotlib.axes.Axes
        Axes object or list of axis objects as return by the matplotlib subplots command
    n_rows : int, optional, default: 1
        The number of rows of the current plot
    n_cols : int, optional, default: 1
        number of columns of the current plot

    Returns
    -------
    type: list
        new 2D list of the axis  which can be referred to as `axis[i_row][j_col]`

    Notes
    -----

    The return axis of the *matplotlib.subplots* command is 2D in the case we have more
    than one row and more than one column. However, if there is only one row or col,
    the list will be 1D, for *n_col = 1* and *n_row = 1* the axis are directly returned.
    This scrips ensures that the list is always started as a 2D list such that the axis
    can be referred to with two indices: *axis[i_row][j_col]*, also if we only have on
    row or columns

    The practical usage of this function is to be able to write more generic code for
    which it is not known before hand how many rows and columns need to be generated
    (it may depend on the user input).

    Examples
    --------

    First create a *matplotlib* plot with only one row of two column

    >>> import matplotlib.pyplot as plt
    >>> from numpy import (sin, cos, pi, linspace)
    >>> fig, axis = plt.subplots(nrows=1, ncols=2)

    At this point, *axis* is a 1D list of 2 *matplotlib.axis.Axes* objects, one for
    each column. So we can plot like this

    >>> x = linspace(0, 2 * pi, num=100)
    >>> l0 = axis[0].plot(x, sin(x))
    >>> l1 = axis[1].plot(x, cos(x))

    Now we want to turn that into a 2D list

    >>> axis_2d = sub_plot_axis_to_2d(axis, n_cols=2)

    Now, axis2d is a 2D list of Axes for 2 columns which are referred as axis[0][0] and
    axis[0][1], so we can plot

    >>> l2d0 = axis_2d[0][0].plot(x, sin(x))
    >>> l2d1 = axis_2d[0][1].plot(x, cos(x))

    Note that we are referring to *axis_2d* as a 2D list with 2 indices in stead of one.
    This allows to create plots on the subplots returned axis in always the same way,
    which is handy if we don't know beforehand how many rows and columns we have.

    """
    new_axis = list()
    if n_cols == 1 and n_rows == 1:
        # We have only one column and one row. Make the new 2D list by appending a list
        # of the axis to the new axis
        new_axis.append([axis])
    elif n_cols == 1 and n_rows > 1:
        for i_r in range(n_rows):
            new_axis.append([axis[i_r]])
    elif n_cols > 1 and n_rows == 1:
        new_axis.append(axis)
    else:
        new_axis = axis

    return new_axis


def set_limits(axis, v_min=None, v_max=None, direction=0):
    """
     Set the x or y limits of the current axis and overwrite them if requested.

    Parameters
    ----------
     v_min: float
        The x or y limit to override if given
     v_max: float
        The x or y maximum to override ig given
    direction: int or str
        Set limit on this axis. 0=x, 1=y, or pass "x" or "y"

    Returns
    -------
    list
        [min, max] with the current x or y limits

    Notes
    -----
    If only the min or max is imposed, obtained the other limit from the current
    settings
    """
    lim = list()
    if direction in (0, "x"):
        current_limits = axis.get_xlim()
    elif direction in (1, "y"):
        current_limits = axis.get_ylim()
    else:
        raise ValueError("Invalid value for the axis parameter")

    if v_min is not None:
        lim.append(v_min)
    else:
        # the minimum is not given, so just return the default value
        lim.append(current_limits[0])
    if v_max is not None:
        lim.append(v_max)
    else:
        # the y maximum is not given, so just return the default value
        lim.append(current_limits[1])

    if direction in (0, "x"):
        axis.set_xlim(lim)
    else:
        axis.set_ylim(lim)

    return lim


def get_valid_legend_handles_labels(axis):
    """
    Function to get all the labels from an axis which are not set to "None"
    (as a string!)

    Parameters
    ----------
    axis: :obj:`SubPlotAxis`

    Returns
    -------
    tuple of list:
        handles, labels

    """
    handles_tpm, labels_tmp = axis.get_legend_handles_labels()
    handles = list()
    labels = list()
    for ii, label in enumerate(labels_tmp):
        if label != "None":
            handles.append(handles_tpm[ii])
            labels.append(label)
    return handles, labels


def flip(items, ncol):
    """
    Flip the legend list to get the lines in horizontal layout

    Parameters
    ----------
    items : list
        list of labels

    ncol : int
        Number of columns


    Returns
    -------
    list
        Reorganised list of labels

    """
    return itertools.chain(*[items[i::ncol] for i in range(ncol)])


def clear_all_legends(axis):
    """
    Remove all the legends from the current axis

    Parameters
    ----------
    axis: list of :obj:AxesSubPlots
    """

    for ax in axis:
        try:
            ax.legend_.remove()
        except AttributeError:
            pass



import numpy as np
from numpy import pi, sqrt, degrees
from numpy.testing import assert_almost_equal

from pymarine.utils.coordinate_transformations import (
    cartesian_to_cylindrical,
    cylindrical_to_cartesian,
    cartesian_to_spherical,
    spherical_to_cartesian,
    latitude_to_colatitude,
    colatitude_to_latitude,
)


def test_cartensian_to_cylindrical():
    # a list of input argument (left) with the expected answer (right)
    input_arguments_and_expected_result = np.array(
        [
            ((0, 0, 0), (0, 0, 0)),
            ((1, 0, 0), (1, 0, 0)),
            ((0, 1, 0), (1, pi / 2, 0)),
            ((0, 0, 1), (0, 0, 1)),
            ((-1, 0, 0), (1, pi, 0)),
            ((0, -1, 0), (1, -pi / 2, 0)),
            ((0, 0, -1), (0, 0, -1)),
            ((-1, 1, 0), (sqrt(2), 3 * pi / 4, 0)),
            ((1, -1, 0), (sqrt(2), -pi / 4, 0)),
            ((0, 1, -1), (1, pi / 2, -1)),
        ]
    )

    # apply all tests as float
    for argument, expected_result in input_arguments_and_expected_result:
        x, y, z = argument[0], argument[1], argument[2]
        result = cartesian_to_cylindrical(x, y, z)
        assert_almost_equal([result], [expected_result])

        # test for degrees as well
        result = cartesian_to_cylindrical(x, y, z, degrees=True)
        expected_result_deg = [
            degrees(t) if i == 1 else t for i, t in enumerate(expected_result)
        ]
        assert_almost_equal([result], [expected_result_deg])

    # test array like result a well
    input_arguments = input_arguments_and_expected_result[
        :, 0, :
    ]  # select the input arguments
    output_arguments = input_arguments_and_expected_result[
        :, 1, :
    ]  # select the output arguments
    xx, yy, zz = input_arguments[:, 0], input_arguments[:, 1], input_arguments[:, 2]
    rr_e, tt_e, zzc_e = (
        output_arguments[:, 0],
        output_arguments[:, 1],
        output_arguments[:, 2],
    )
    rr, tt, zzc = cartesian_to_cylindrical(xx, yy, zz)
    assert_almost_equal([rr], [rr_e])
    assert_almost_equal([tt], [tt_e])
    assert_almost_equal([zzc], [zzc_e])

    rr, tt, zzc = cartesian_to_cylindrical(xx, yy, zz, degrees=True)
    assert_almost_equal([rr], [rr_e])
    assert_almost_equal([tt], [np.rad2deg(tt_e)])
    assert_almost_equal([zzc], [zzc_e])


def test_cylindrical_to_cartesian():
    # a list of input argument (left) with the expected answer (right)
    input_arguments_and_expected_result = np.array(
        [
            ((0, 0, 0), (0, 0, 0)),
            ((1, 0, 0), (1, 0, 0)),
            ((1, pi / 2, 0), (0, 1, 0)),
            ((0, 0, 1), (0, 0, 1)),
            ((1, pi, 0), (-1, 0, 0)),
            ((1, -pi / 2, 0), (0, -1, 0)),
            ((0, 0, -1), (0, 0, -1)),
            ((sqrt(2), 3 * pi / 4, 0), (-1, 1, 0)),
            ((sqrt(2), -pi / 4, 0), (1, -1, 0)),
            ((1, pi / 2, -1), (0, 1, -1)),
        ]
    )

    # apply all tests
    for argument, expected_result in input_arguments_and_expected_result:
        result = cylindrical_to_cartesian(argument[0], argument[1], argument[2])
        assert_almost_equal([result], [expected_result])

        # test for degrees as well
        argument_deg = [degrees(t) if i == 1 else t for i, t in enumerate(argument)]
        result = cylindrical_to_cartesian(
            argument_deg[0], argument_deg[1], argument_deg[2], degrees=True
        )
        assert_almost_equal([result], [expected_result])

    # test array like result a well
    input_arguments = input_arguments_and_expected_result[
        :, 0, :
    ]  # select the input arguments
    output_arguments = input_arguments_and_expected_result[
        :, 1, :
    ]  # select the output arguments
    rr, tt, zzc = input_arguments[:, 0], input_arguments[:, 1], input_arguments[:, 2]
    xx_e, yy_e, zz_e = (
        output_arguments[:, 0],
        output_arguments[:, 1],
        output_arguments[:, 2],
    )
    xx, yy, zz = cylindrical_to_cartesian(rr, tt, zzc)
    assert_almost_equal([xx], [xx_e])
    assert_almost_equal([yy], [yy_e])
    assert_almost_equal([zz], [zz_e])

    xx, yy, zz = cylindrical_to_cartesian(rr, np.rad2deg(tt), zzc, degrees=True)
    assert_almost_equal([xx], [xx_e])
    assert_almost_equal([yy], [yy_e])
    assert_almost_equal([zz], [zz_e])


def test_cartesian_to_spherical():
    # a list of input argument (left) with the expected answer (right)
    input_arguments_and_expected_result = np.array(
        [
            [[0, 0, 0], [0, 0, 0]],
            [[1, 0, 0], [1, pi / 2, 0]],
            [[0, 1, 0], [1, pi / 2, pi / 2]],
            [[0, 0, 1], [1, 0, 0]],
            [[-1, 0, 0], [1, pi / 2, pi]],
            [[0, -1, 0], [1, pi / 2, -pi / 2]],
            [[0, 0, -1], [1, pi, 0]],
            [[-1, 1, 0], [sqrt(2), pi / 2, 3 * pi / 4]],
            [[1, -1, 0], [sqrt(2), pi / 2, -pi / 4]],
            [[0, 1, -1], [sqrt(2), 3 * pi / 4, pi / 2]],
        ]
    )

    # apply all tests
    for argument, expected_result in input_arguments_and_expected_result:
        result = cartesian_to_spherical(argument[0], argument[1], argument[2])
        assert_almost_equal([result], [expected_result])

        # test for degrees as well
        result = cartesian_to_spherical(
            argument[0], argument[1], argument[2], degrees=True
        )
        expected_result_deg = [
            degrees(t) if i > 0 else t for i, t in enumerate(expected_result)
        ]
        assert_almost_equal([result], [expected_result_deg])

    # test array like result a well
    input_arguments = input_arguments_and_expected_result[
        :, 0, :
    ]  # select the input arguments
    output_arguments = input_arguments_and_expected_result[
        :, 1, :
    ]  # select the output arguments
    xx, yy, zz = input_arguments[:, 0], input_arguments[:, 1], input_arguments[:, 2]
    rr_e, tt_e, pp_e = (
        output_arguments[:, 0],
        output_arguments[:, 1],
        output_arguments[:, 2],
    )
    rr, tt, pp = cartesian_to_spherical(xx, yy, zz)
    assert_almost_equal([rr], [rr_e])
    assert_almost_equal([tt], [tt_e])
    assert_almost_equal([pp], [pp_e])

    rr, tt, pp = cartesian_to_spherical(xx, yy, zz, degrees=True)
    assert_almost_equal([rr], [rr_e])
    assert_almost_equal([tt], [np.rad2deg(tt_e)])
    assert_almost_equal([pp], [np.rad2deg(pp_e)])


def test_spherical_to_cartesian():
    # a list of input argument (left) with the expected answer (right)
    input_arguments_and_expected_result = np.array(
        [
            ((0, 0, 0), (0, 0, 0)),
            ((1, pi / 2, 0), (1, 0, 0)),
            ((1, pi / 2, pi / 2), (0, 1, 0)),
            ((1, 0, 0), (0, 0, 1)),
            ((1, pi / 2, pi), (-1, 0, 0)),
            ((1, pi / 2, -pi / 2), (0, -1, 0)),
            ((1, pi, 0), (0, 0, -1)),
            ((sqrt(2), pi / 2, 3 * pi / 4), (-1, 1, 0)),
            ((sqrt(2), pi / 2, -pi / 4), (1, -1, 0)),
            ((sqrt(2), 3 * pi / 4, pi / 2), (0, 1, -1)),
        ]
    )

    # apply all tests
    for argument, expected_result in input_arguments_and_expected_result:
        result = spherical_to_cartesian(argument[0], argument[1], argument[2])
        assert_almost_equal([result], [expected_result])

        # test for degrees as well
        argument_deg = [degrees(t) if i > 0 else t for i, t in enumerate(argument)]
        result = spherical_to_cartesian(
            argument_deg[0], argument_deg[1], argument_deg[2], degrees=True
        )
        assert_almost_equal([result], [expected_result])

    # test array like result a well
    input_arguments = input_arguments_and_expected_result[
        :, 0, :
    ]  # select the input arguments
    output_arguments = input_arguments_and_expected_result[
        :, 1, :
    ]  # select the output arguments
    rr, tt, pp = input_arguments[:, 0], input_arguments[:, 1], input_arguments[:, 2]
    xx_e, yy_e, zz_e = (
        output_arguments[:, 0],
        output_arguments[:, 1],
        output_arguments[:, 2],
    )
    xx, yy, zz = spherical_to_cartesian(rr, tt, pp)
    assert_almost_equal([xx], [xx_e])
    assert_almost_equal([yy], [yy_e])
    assert_almost_equal([zz], [zz_e])

    xx, yy, zz = spherical_to_cartesian(
        rr, np.rad2deg(tt), np.rad2deg(pp), degrees=True
    )
    assert_almost_equal([xx], [xx_e])
    assert_almost_equal([yy], [yy_e])
    assert_almost_equal([zz], [zz_e])


def test_latitude_to_colatitude():
    input_arguments_and_expected_result = np.array(
        [
            (0, pi / 2),
            (pi / 4, pi / 4),
            (pi / 2, 0),
        ]
    )

    for argument, expected_result in input_arguments_and_expected_result:
        result = latitude_to_colatitude(argument)
        assert_almost_equal([result], [expected_result])

        argument_deg = degrees(argument)
        result = latitude_to_colatitude(argument_deg, degrees=True)
        expected_result_deg = degrees(expected_result)
        assert_almost_equal([result], [expected_result_deg])

    # test the array input as well
    ll = input_arguments_and_expected_result[:, 0]
    cl_e = input_arguments_and_expected_result[:, 1]

    cl = latitude_to_colatitude(ll)
    assert_almost_equal([cl], [cl_e])

    cl = latitude_to_colatitude(np.rad2deg(ll), degrees=True)
    assert_almost_equal([cl], [np.rad2deg(cl_e)])


def test_colatitude_to_latitude():
    input_arguments_and_expected_result = np.array(
        [
            (0, pi / 2),
            (pi / 4, pi / 4),
            (pi / 2, 0),
        ]
    )

    for argument, expected_result in input_arguments_and_expected_result:
        result = colatitude_to_latitude(argument)
        assert_almost_equal([result], [expected_result])

        argument_deg = degrees(argument)
        result = colatitude_to_latitude(argument_deg, degrees=True)
        expected_result_deg = degrees(expected_result)
        assert_almost_equal([result], [expected_result_deg])

    # test the array input as well
    cl = input_arguments_and_expected_result[:, 0]
    ll_e = input_arguments_and_expected_result[:, 1]

    ll = latitude_to_colatitude(cl)
    assert_almost_equal([ll], [ll_e])

    ll = latitude_to_colatitude(np.rad2deg(cl), degrees=True)
    assert_almost_equal([ll], [np.rad2deg(ll_e)])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from numpy.testing import (assert_almost_equal)

from pymarine.waves.wave_fields import (Wave1D, Wave2D)


def test_wave_1d():
    hs_in = 3.0

    for wave_construction in ("FFT", "DFTpolar", "DFTcartesian"):
        if wave_construction == "DFTpolar":
            factor = 1.0
        else:
            factor = 0.5

        wave1d = Wave1D(Hs=hs_in, n_kx_nodes=1024, Lx=2000, nx_points=1024,
                        wave_construction=wave_construction)

        # check the energy. Note that for FFT and DFTcartesian we have mirrored the spectrum, and we
        # have every bin double; therefore factor = 0.5 to take only half of the domain energy
        hs_out = 4 * np.sqrt(wave1d.delta_kx * np.sum(wave1d.spectrumK) * factor)
        assert_almost_equal(hs_in, hs_out, decimal=1)

        # bring the delta_omega inside as it is not equidistant anymore
        hs_out = 4 * np.sqrt(np.sum(wave1d.delta_omega * wave1d.spectrumW) * factor)
        assert_almost_equal(hs_in, hs_out, decimal=1)


def test_wave_2d():
    hs_in = 3.0

    for wave_construction in ("FFT", "DFTpolar", "DFTcartesian"):

        wave1d = Wave1D(Hs=hs_in, n_kx_nodes=128, Lx=1000, nx_points=128,
                        wave_construction=wave_construction)

        wave2d = Wave2D(wave1D=wave1d)

        # check the significant wave heigt from the amplitude
        hs_out = 4 * wave2d.amplitude.std()
        assert_almost_equal(hs_in, hs_out, decimal=1)



"""
Several date/time helper functions.
"""

import argparse
import datetime
from time import strptime

import numpy as np


def _from_matlab_to_python_num_days(x):
    # Matlab epoch time start at the year 0, python at the 1-1-0001.
    # Both formats differ 366 days.
    # So to convert the matlab format to the python format, convert it here
    return (
        datetime.datetime.fromordinal(int(x))
        + datetime.timedelta(days=x % 1)
        - datetime.timedelta(days=366)
    )


_from_matlab_to_python_num_days_vectorized = np.vectorize(
    _from_matlab_to_python_num_days
)


def matlabnum2date(x):
    """Convert a Matlab numerical date/time representation into a Python datetime object

    Parameters
    ----------

    Parameters
    ----------
    x : float or ndarrray

        Matlab numerical date/time representation giving the number of days since
        0000 00 00:00:00

    Returns
    -------
    number_of_days : ndarray or scalar of the type datetime`
        Date/time corresponding to the float value `x`

    Notes
    -----
    * In Matlab, the numerical date/time representation gives the number of days since
      Jan 00 0000 00:00:00
    * In Python, the numerical date/time representation gives the number of days since
      Jan 01 0001 00:00:00
    * To convert from matlab to python, 366 days need to be subtracted from the matlab
      numerical date/time representation to get the Python numerical data/time
      representation (note that matlab starts at Jan 0th!)

    Examples
    --------

    The following two dates in numerical representation are obtained using the Matlab
    `datenum` function

    * 2012-12-21T12:12:12 -> 7.352245084722222e+05 # days since 0000 00 00:00:00
    * 1973-11-12T09:15:43 -> 7.209403859143519e+05 # days since 0000 00 00:00:00

    >>> num_date_matlab = np.array([7.352245084722222222222222e+05,
    ...                             7.209403859143519e+05])

    >>> matlabnum2date(num_date_matlab)
    array([datetime.datetime(2012, 12, 21, 12, 12, 12, 3),
           datetime.datetime(1973, 11, 12, 9, 15, 43, 5)], dtype=object)

    """
    try:
        number_of_days = _from_matlab_to_python_num_days(x)
    except TypeError:
        x = np.asarray(x)
        if not x.size:
            raise
        number_of_days = _from_matlab_to_python_num_days_vectorized(x)

    return number_of_days


def valid_date(s):
    """Check if supplied data *s* is a valid date for the format Year-Month-Day

    Parameters
    ----------
    s : str
        A valid date in the form of YYYY-MM-DD, so first the year, then the month, then
        the day

    Returns
    -------
    datetime
        Date object with the year, month, day obtained from the valid string
        representation

    Raises
    ------
    argparse.ArgumentTypeError:

    Notes
    -----
    This is a helper function for the argument parser module argparse which allows you
    to check if the argument passed on the command line is a valid date.

    Examples
    --------

    This is the direct usage of `valid_date` to see if the date supplied is of format
    YYYY-MM-DD

    >>> try:
    ...     date = valid_date("1973-11-12")
    ... except argparse.ArgumentTypeError:
    ...     print("This date is invalid")
    ... else:
    ...     print("This date is valid")
    This date is valid

    In case an invalid date is supplied

    >>> try:
    ...     date = valid_date("1973-15-12")
    ... except argparse.ArgumentTypeError:
    ...     print("This date is invalid")
    ... else:
    ...     print("This date is valid")
    This date is invalid


    Here it is demonstrated how to add a '--startdate' command line option to the
    argparse parser which checks if a valid date is supplied

    >>> parser = argparse.ArgumentParser()
    >>> p = parser.add_argument("--startdate",
    ...                         help="The Start Date - format YYYY-MM-DD ",
    ...                         required=True,
    ...                         type=valid_date)

    References
    ----------

    https://stackoverflow.com/questions/25470844/
        specify-format-for-input-arguments-argparse-python
    """

    try:
        return strptime(s, "%Y-%m-%d")
    except ValueError:
        msg = f"Not a valid date: '{s}'.\nSupply date as YYYY-MM-DD"
        raise argparse.ArgumentTypeError(msg)



import logging

import matplotlib.pyplot as plt

import pymarine.waves.wave_fields as wf
from pymarine.utils.misc import Timer

logging.basicConfig(level=logging.INFO)

# spectrum_type = "gauss"
spectrum_type = "jonswap"

wave1 = wf.Wave1D(
    delta_t=1,
    spectrum_type=spectrum_type,
    xmax=3000,
    n_kx_nodes=512,
    kx_max=3.14,
    nt_samples=1000,
    nx_points=512,
    wave_construction="DFTpolar",
)
wave2 = wf.Wave1D(
    delta_t=1,
    spectrum_type=spectrum_type,
    xmax=3000,
    wave_selection="EqualEnergyBins",
    lock_nodes_to_wave_one=False,
    use_subrange_energy_limits=False,
    n_bins_equal_energy=20,
    kx_max=3.14,
    nt_samples=1000,
    nx_points=512,
    wave_construction="DFTpolar",
)

wave3 = wf.Wave1D(
    delta_t=1,
    spectrum_type=spectrum_type,
    xmax=3000,
    n_kx_nodes=1024,
    kx_max=3.14,
    nt_samples=1000,
    nx_points=1024,
    wave_construction="FFT",
)

wave1.make_report()
wave2.make_report()
wave3.make_report()

fig, axis = wave1.plot_spectrum(
    add_limit_markers=True, add_hs_estimate=True, plot_title="Compare " "Spectra"
)
fig, axis = wave2.plot_spectrum(
    fig=fig,
    axis=axis,
    linecolor="r",
    linestyle="--",
    line_markers="v",
    y_n_points_label=0.7,
    markersize=10,
)

fig, axis = wave1.plot_wave(plot_title="Wave compare")
fig, axis = wave2.plot_wave(fig=fig, ax=axis, linestyle="--")
fig, axis = wave3.plot_wave(fig=fig, ax=axis, linestyle="-.")

with Timer(message=wave1.name) as t:
    while wave1.time < wave1.t_end:
        wave1.propagate_wave()

with Timer(message=wave2.name) as t:
    while wave2.time < wave2.t_end:
        wave2.propagate_wave()

with Timer(message=wave3.name) as t:
    while wave3.time < wave3.t_end:
        wave3.propagate_wave()

plt.show()

wave1.reset_time(nt_samples=100, delta_t=0.1)
# wave2.reset_time(nt_samples=1000, delta_t=0.1)
# wave3.reset_time(nt_samples=1000, delta_t=0.1)

# movie = wave1.animate_wave(interval=1, y_min=-1, y_max=1, figsize=(12,4))
# plt.show()

# movie = wave2.animate_wave(interval=1, y_min=-1, y_max=1, figsize=(12,4))
# plt.show()
#
# movie = wave3.animate_wave(interval=1, y_min=-1, y_max=1, figsize=(12,4))
# plt.show()



"""
Definition of some common coordinate system conversions

Notes
-----
* The coordinate transformations defined in this section are a selection of the
  coordinate_transformation implemented in the astropysics model. The reason why this module is not
  installed by it self is that this module has not been updated for Python 3.
  It is therefore easier just to take the part of the module which is relevant for HMC

* The definitions of the spherical coordinates follow the physics convention as described here_

References
----------

* https://en.wikipedia.org/wiki/Spherical_coordinate_system

.. _here:
    https://en.wikipedia.org/wiki/Spherical_coordinate_system

"""

import numpy as np
from numpy import pi


# <--------------------Functional coordinate transforms------------------------->
def cartesian_to_polar(x, y, degrees=False):
    """Converts arrays in 2D rectangular Cartesian coordinates to polar
    coordinates.

    Parameters
    ----------
    x : float or array_like
        First cartesian coordinate
    y : float or array_like
        Second cartesian coordinate
    degrees : boolean
        If True, the output theta angle will be in degrees, otherwise radians. (Default = False)

    Returns
    -------
    tuple of 2 floats or ndarrays
        Polar coordinates (r, theta) where theta is measured from the +x axis increasing towards
        the +y axis

    """
    npx, npy = np.array(x), np.array(y)
    r = (npx * npx + npy * npy) ** 0.5
    t = np.arctan2(npy, npx)
    if degrees:
        t = np.degrees(t)

    return r, t


def polar_to_cartesian(r, t, degrees=False):
    """Converts arrays in 2D polar coordinates to rectangular cartesian
    coordinates.

    Parameters
    ----------
    r : float or array_like
        Radial coordinate
    t : float or array_like
        Azimuthal angle from +x-axis increasing towards +y-axis
    degrees : boolean
        If True, the input angles will be in degrees, otherwise radians. (Default value = False)

    Returns
    -------
    tuple of 2 floats or ndarrays
        Cartesian coordinates (x,y)

    """
    npt, npr = np.array(t), np.array(r)
    if degrees:
        npt = np.radians(npt)

    return npr * np.cos(npt), r * np.sin(npt)


def cartesian_to_spherical(x, y, z, degrees=False):
    """Converts three arrays in 3D rectangular cartesian coordinates to
    spherical polar coordinates.

    Note that the spherical coordinates are in *physicist* convention such that (1,pi/2,0) is
    x-axis.

    Parameters
    ----------
    x : float or array_like
        First cartesian coordinate
    y : float or array_like
        Second cartesian coordinate
    z : float or array_like
        Third cartesian coordinate
    degrees : boolean
        If True, the output theta angle will be in degrees, otherwise radians. (Default = False)

    Returns
    -------
    tuple of 3 floats or ndarrays
        Spherical coordinates (r,theta,phi)

    Examples
    --------
    Convert the coordinates (1, 0, 0)

    >>> radius, theta, phi = cartesian_to_spherical(1, 0, 0)
    >>> print("Radius = {:.4g}  Theta = {:.4g}  Phi = {:.4g}".format(radius, theta, phi))
    Radius = 1  Theta = 1.571  Phi = 0

    >>> radius, theta, phi = cartesian_to_spherical(1, 1, 1, degrees=True)
    >>> print("Radius = {:.4g}  Theta = {:.4g}  Phi = {:.4g}".format(radius, theta, phi))
    Radius = 1.732  Theta = 54.74  Phi = 45


    Convert array_like coordinates

    >>> xx, yy, zz = [0, 0, 1 ], [1, 0, 1 ], [0, 1, 1 ]
    >>> radii, thetas, phis = cartesian_to_spherical(xx, yy, zz)
    >>> for ii, radius in enumerate(radii):
    ...   print("({:.2f}, {:.2f} {:.2f}) -> ({:.4f}, {:.4f}, {:.4f})"
    ...         "".format(xx[ii], yy[ii], zz[ii], radius, thetas[ii], phis[ii]))
    (0.00, 1.00 0.00) -> (1.0000, 1.5708, 1.5708)
    (0.00, 0.00 1.00) -> (1.0000, 0.0000, 0.0000)
    (1.00, 1.00 1.00) -> (1.7321, 0.9553, 0.7854)

    References
    ----------
    https://en.wikipedia.org/wiki/Spherical_coordinate_system

    """
    npx, npy, npz = np.array(x), np.array(y), np.array(z)
    xsq, ysq, zsq = npx * npx, npy * npy, npz * npz
    r = (xsq + ysq + zsq) ** 0.5
    t = np.arctan2((xsq + ysq) ** 0.5, z)
    p = np.arctan2(npy, npx)
    if degrees:
        t, p = np.degrees(t), np.degrees(p)
    return r, t, p


def spherical_to_cartesian(r, t, p, degrees=False):
    """Converts arrays in 3D spherical polar coordinates to rectangular cartesian coordinates.

    Parameters
    ----------
    r : float or array like
        Radial coordinate
    t : float or array_like
        Colatitude (angle from z-axis)
    p : float or array_lie
        Azimuthal angle from +x-axis increasing towards +y-axis
    degrees : boolean
        If True, the input angles will be in degrees, otherwise radians. (Default value = False)

    Returns
    -------
    tuple of 3 floats or ndarrays
        Cartesian coordinates (x,y,z)

    Notes
    -----
    * The spherical coordinates are in *physicist* convention such that (1,pi/2,0) is the x-axis.

    Examples
    --------
    >>> xx, yy, zz = spherical_to_cartesian(1, 0, 0)
    >>> print("xx = {:.4g}  yy = {:.4g}  zz = {:.4g}".format(xx, yy, zz))
    xx = 0  yy = 0  zz = 1
    >>> xx, yy, zz = spherical_to_cartesian(1.7321, 54.736, 45, degrees=True)
    >>> print("xx = {:.4g}  yy = {:.4g}  zz = {:.4g}".format(xx, yy, zz))
    xx = 1  yy = 1  zz = 1
    >>> rr, tt, pp = np.array([1, 2, 3 ]), np.array([0, 45, 90 ]), np.array([0, -45, 45 ])
    >>> xx, yy, zz = spherical_to_cartesian(rr, tt, pp, degrees=True)
    >>> for ii, x in enumerate(xx):
    ...   print("({:6.2f}, {:6.2f} {:6.2f}) -> ({:6.2f}, {:6.2f}, {:6.2f})"
    ...         "".format(rr[ii], tt[ii], pp[ii], x, yy[ii], zz[ii]))
    (  1.00,   0.00   0.00) -> (  0.00,   0.00,   1.00)
    (  2.00,  45.00 -45.00) -> (  1.00,  -1.00,   1.41)
    (  3.00,  90.00  45.00) -> (  2.12,   2.12,   0.00)
    """
    npr, npt, npp = np.array(r), np.array(t), np.array(p)
    if degrees:
        npt, npp = np.radians(npt), np.radians(npp)
    x = npr * np.sin(npt) * np.cos(npp)
    y = npr * np.sin(npt) * np.sin(npp)
    z = npr * np.cos(npt)

    return x, y, z


def latitude_to_colatitude(lat, degrees=False):
    """Converts from latitude  (i.e. 0 at the equator) to colatitude/inclination
    (i.e. "theta" in physicist convention).

    Parameters
    ----------
    lat : float or array_like
        Latitude

    degrees : bool
         If True, use degrees (Default value = False)

    Returns
    -------
    float or ndarray
        Colatitude

    Examples
    --------
    >>> latitude_to_colatitude(10, degrees=True)
    80
    >>> phi_list = [np.pi / i for i in range(1,6)]
    >>> co_latitude = latitude_to_colatitude(phi_list)
    >>> for ii, phi in enumerate(phi_list):
    ...    print("{:6.4f} -> {:7.4f}".format(phi, co_latitude[ii]))
    3.1416 -> -1.5708
    1.5708 ->  0.0000
    1.0472 ->  0.5236
    0.7854 ->  0.7854
    0.6283 ->  0.9425
    """
    lat_array = np.array(lat)
    if degrees:
        return 90 - lat_array
    else:
        return pi / 2 - lat_array


def colatitude_to_latitude(theta, degrees=False):
    """Converts from colatitude/inclination (i.e. "theta" in physicist convention)
    to latitude (i.e. 0 at the equator).

    Parameters
    ----------
    theta : float or array-like
        input colatitude
    degrees : bool
        If True, the input is interpreted as degrees, otherwise radians. (Default value = False)

    Returns
    -------
    float or ndarray
        Latitude

    Examples
    --------
    >>> colatitude_to_latitude(15, degrees=True)
    75
    >>> phi_list = [np.pi / i for i in range(1,6)]
    >>> latitude = colatitude_to_latitude(phi_list)
    >>> for ii, phi in enumerate(phi_list):
    ...    print("{:6.4f} -> {:7.4f}".format(phi, latitude[ii]))
    3.1416 -> -1.5708
    1.5708 ->  0.0000
    1.0472 ->  0.5236
    0.7854 ->  0.7854
    0.6283 ->  0.9425


    """
    theta_array = np.array(theta)
    if degrees:
        return 90 - theta_array
    else:
        return pi / 2 - theta_array


def cartesian_to_cylindrical(x, y, z, degrees=False):
    """Converts three arrays in 3D rectangular Cartesian coordinates to cylindrical
    polar coordinates.

    Parameters
    ----------
    x : float or array-like
        x cartesian coordinate
    y : float or array-like
        y cartesian coordinate
    z : float or array-like
        z cartesian coordinate
    degrees : bool
        If True, the output angles will be in degrees, otherwise radians. (Default value = False)

    Returns
    -------
    tuple of 3 floats or ndarrays
        Cylindrical coordinates as a (rho,theta,z),  with theta increasing from +x to +y and
        theta = 0 at x-axis.

    """
    s, t = cartesian_to_polar(x, y, degrees)
    return s, t, z


def cylindrical_to_cartesian(s, t, z, degrees=False):
    """Converts three arrays in cylindrical polar coordinates to 3D rectangular
    Cartesian coordinates.

    Parameters
    ----------
    s : float or array-like
        radial polar coordinate
    t : float or array-like
        polar angle (increasing from +x to +y, 0 at x-axis)
    z : float or array-like
        z coordinate
    degrees : bool
        If True, the input angle will be in degrees, otherwise radians. (Default value = False)

    Returns
    -------
    tuple of 3 floats or ndarrays
        Cartesian coordinates as an (x,y,z) tuple.

    """
    x, y = polar_to_cartesian(s, t, degrees)
    return x, y, z



#!/usr/bin/env python

import os
import re
import sys
import time

from numpy import array
from numpy.testing import (
    assert_almost_equal,
    assert_equal,
    assert_raises,
    assert_string_equal,
)

try:
    from pymarine.utils import Q_
except ImportError:
    Q_ = None
from pymarine.utils.misc import (
    Chdir,
    Timer,
    clear_argument_list,
    clear_path,
    get_clean_version,
    get_python_version_number,
    get_regex_pattern,
    get_value_magnitude,
    is_exe,
    set_default_dimension,
)


def test_timer():
    number_of_seconds = 1.0
    with Timer() as timer:
        time.sleep(number_of_seconds)
    assert_almost_equal([timer.secs], [number_of_seconds], decimal=0)


def test_chdir():
    python_dir = os.path.dirname(sys.prefix)
    with Chdir(new_path=python_dir):
        current_dir = os.getcwd()

    assert_string_equal(python_dir, current_dir)


def test_is_exe():
    # works only on linux
    test_file = "ls"
    this_is_an_executable = is_exe(test_file)
    if os.name != "nt":
        # this test can only succeed on linux
        assert_equal(this_is_an_executable, True)


def test_clean_path():
    python_dir = os.path.dirname(sys.prefix)
    new_path = clear_path(python_dir)
    assert_string_equal(python_dir, new_path)


def test_get_clean_version():
    version_number = "1.1"
    test_version_clean = "test_clean_version-" + version_number
    test_version_string = test_version_clean + "+jkjfds"
    clean_version = get_clean_version(test_version_string)
    assert_string_equal(version_number, clean_version)


def test_get_python_version_number():
    get_python_version_number(sys.version_info)


def test_get_regex_pattern():
    regular_expression = r"\s\d+\w"
    result_1 = get_regex_pattern(regular_expression)
    result_2 = re.compile(f"{regular_expression}")
    assert_equal(result_1, result_2)


def test_clear_argument_list():
    arg_old = ["blabla\r", "fdsafds", "blablabla\r", "blabla\n"]
    arg_new = clear_argument_list(arg_old)
    arg_new_expected = ["blabla", "fdsafds", "blablabla", "blabla\n"]
    assert_equal(arg_new, arg_new_expected)


def test_set_default_dimensions():
    # in case no dimension is given, we add the default dimension
    val_in = "1.0"
    val_out = set_default_dimension(val_in, "meter")
    assert_equal(val_out, Q_(val_in, "meter"))

    # in case x dimension is given, we do not the default dimension
    val_in = Q_(2.0, "meter")
    val_out = set_default_dimension(val_in, "meter")
    assert_equal(val_out, Q_(val_in, "meter"))

    # in case we pass a value with a different dimensionality as the default dimension,
    # raise an AssertionError
    val_in = Q_(3.0, "second")
    assert_raises(AssertionError, set_default_dimension, val_in, "meter")

    val_in = [0, 1, 2]
    val_out = set_default_dimension(val_in, "meter")
    for i, x in enumerate(val_out):
        assert_equal(x, Q_(val_in[i], "meter"))

    val_in = ["0", "1", "2"]
    val_out = set_default_dimension(val_in, "meter")
    for i, x in enumerate(val_out):
        assert_equal(x, Q_(val_in[i], "meter"))

    val_in = array([0, 1, 2])
    val_out = set_default_dimension(val_in, "meter")
    for i, x in enumerate(val_out):
        assert_equal(x, Q_(val_in[i], "meter"))
    assert_equal(val_out.all(), Q_(val_in, "meter").all())

    val_in = Q_(array([0, 1, 2]), "meter")
    val_out = set_default_dimension(val_in, "meter")
    for i, x in enumerate(val_out):
        assert_equal(x, Q_(val_in[i], "meter"))
    assert_equal(val_out.all(), Q_(val_in, "meter").all())

    # We pass an array with dimension second while we request a default dimension of
    # meter. This should raise an error
    val_in = Q_(array([0, 1, 2]), "second")
    assert_raises(AssertionError, set_default_dimension, val_in, "meter")

    val_in = "1.0 meter"
    val_out = set_default_dimension(val_in, "meter")
    assert_equal(val_out, Q_("1 meter"))


def test_value_magnitude():
    frequency_bare = 1.0
    frequency_unit = Q_(frequency_bare, "Hz")

    frequency_bare_2 = get_value_magnitude(frequency_unit)

    assert_equal([frequency_bare], [frequency_bare_2])

    frequency_bare_2 = get_value_magnitude(frequency_bare)
    assert_equal([frequency_bare], [frequency_bare_2])

    frequency_none = get_value_magnitude(None)
    assert_equal([frequency_none], [None])

    velocity_knots_mag = get_value_magnitude(Q_("1.0 knots"))
    assert_almost_equal([velocity_knots_mag], [0.514444444])

    velocity_knots_mag = get_value_magnitude(
        Q_("1.0 knots"), convert_to_base_units=False
    )
    assert_almost_equal([velocity_knots_mag], [1.0])



#!/usr/bin/env python
import numpy as np
from numpy import pi
from numpy.testing import assert_almost_equal, assert_equal

from pymarine.waves.wave_spectra import (
    alpha_jonswap,
    d_omega_e_prime,
    initialize_phase,
    mask_out_of_range,
    omega_critical,
    omega_deep_water,
    omega_e_vs_omega,
    omega_peak_jonswap,
    omega_vs_omega_e,
    rayleigh_cdf,
    rayleigh_pdf,
    set_heading,
    specspecs,
    spectrum2d_complex_amplitudes,
    spectrum2d_to_spectrum2d_encountered,
    spectrum_gauss,
    spectrum_jonswap,
    spectrum_jonswap_k_domain_2,
    spectrum_to_complex_amplitudes,
    spectrum_to_spectrum_encountered,
    spectrum_wave_k_domain,
    spreading_function,
    spreading_function2,
    thetaspreadspecs,
)


def test_omega_peak_jonswap():
    peak_expected = 4.678462267825839
    peak = omega_peak_jonswap(wind_speed=10, fetch=1000)
    assert_almost_equal([peak], [peak_expected])


def test_alpha_jonswap():
    alpha_expected = 0.0277127
    alpha = alpha_jonswap(wind_speed=10, fetch=1000)
    assert_almost_equal([alpha], [alpha_expected])


def test_spreading_function2():
    n_size = 10
    directions = np.linspace(0, 360, n_size, endpoint=False)
    result = spreading_function2(theta=directions, theta0=100, s_spreading_factor=5)
    result_expected = np.array(
        [
            4.5273054e-01,
            1.0557061e-01,
            1.4777023e-09,
            9.2072475e-03,
            6.4647155e-01,
            1.1260938e-02,
            3.7390244e-10,
            9.3699574e-02,
            4.7425908e-01,
            2.5123128e-04,
        ]
    )
    assert_almost_equal(result, result_expected)


def test_spreading_function():
    n_size = 10
    directions = np.linspace(0, 2 * pi, n_size, endpoint=False)
    result_expected = np.array(
        [0.6366198, 0.4166731, 0.0607918, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0607918, 0.4166731]
    )
    result = spreading_function(theta=directions, theta0=0, n_spreading_factor=2)
    assert_almost_equal(result, result_expected)


def test_spectrum_gauss():
    n_size = 10
    frequencies = np.linspace(0, 2.5, n_size)
    result = spectrum_gauss(frequencies, Hs=2, Tp=10, sigma=2.4, spectral_version="dnv")
    result_expected = np.array(
        [
            0.0401565,
            0.0411156,
            0.0415374,
            0.0414051,
            0.0407241,
            0.0395213,
            0.0378436,
            0.035755,
            0.0333321,
            0.0306599,
        ]
    )
    assert_almost_equal(result, result_expected)

    result = spectrum_gauss(frequencies, Hs=2, Tp=10, spectral_version="sim")
    result_expected = np.array(
        [
            3.0615835e-022,
            2.7659152e-007,
            8.1180934e-001,
            7.7409033e-003,
            2.3980138e-013,
            2.4134267e-032,
            7.8911262e-060,
            8.3823596e-096,
            2.8927840e-140,
            3.2433077e-193,
        ]
    )
    assert_almost_equal(result, result_expected)

    n_size = 10000
    frequencies = np.linspace(0, 10, n_size)
    delta_f = np.diff(frequencies)[0]
    hs_values = np.linspace(0, 10, num=21, endpoint=True)
    for hs_in in hs_values:
        # check version 0
        result = spectrum_gauss(
            omega=frequencies, Hs=hs_in, Tp=10, spectral_version="dnv"
        )
        hs_out = 4 * np.sqrt(result.sum() * delta_f)
        assert_almost_equal(hs_out, hs_in)

        # check version 1
        result = spectrum_gauss(
            omega=frequencies, Hs=hs_in, Tp=10, spectral_version="sim"
        )
        hs_out = 4 * np.sqrt(result.sum() * delta_f)
        assert_almost_equal(hs_out, hs_in)

    hs_in = 3
    tp_values = np.linspace(1.0, 20, num=21, endpoint=True)
    sigma_values = np.linspace(0.01, 0.1, num=5, endpoint=True)
    for tp in tp_values:
        # check version 0
        for sigma in sigma_values:
            result = spectrum_gauss(
                omega=frequencies, Hs=hs_in, Tp=tp, sigma=sigma, spectral_version="dnv"
            )
            hs_out = 4 * np.sqrt(result.sum() * delta_f)
            assert_almost_equal(hs_out, hs_in, decimal=1)

        # check version 1
        result = spectrum_gauss(
            omega=frequencies, Hs=hs_in, Tp=tp, spectral_version="sim"
        )
        hs_out = 4 * np.sqrt(result.sum() * delta_f)
        assert_almost_equal(hs_out, hs_in, decimal=3)


def test_spectrum_jonswap():
    n_size = 10
    frequencies = np.linspace(0, 2.5, n_size)
    result = spectrum_jonswap(
        omega=frequencies, Hs=3.0, Tp=11, gamma=3.5, spectral_version="dnv"
    )
    result_expected = np.array(
        [
            0.0000000e00,
            2.2801231e-08,
            2.8604450e00,
            3.6208292e-01,
            1.0375968e-01,
            3.5798406e-02,
            1.4655414e-02,
            6.8345663e-03,
            3.5190394e-03,
            1.9568243e-03,
        ]
    )
    assert_almost_equal(result, result_expected)

    # test the other version as well
    result = spectrum_jonswap(
        omega=frequencies, Hs=3, Tp=11, gamma=3.5, spectral_version="sim"
    )
    result_expected = np.array(
        [
            0.0000000e00,
            2.2749758e-08,
            2.8539876e00,
            3.6126553e-01,
            1.0352545e-01,
            3.5717592e-02,
            1.4622330e-02,
            6.8191375e-03,
            3.5110953e-03,
            1.9524068e-03,
        ]
    )
    assert_almost_equal(result, result_expected)

    # jonswap 2 should give the same result as jonswap with 'version=1' flag on
    result2 = spectrum_jonswap(
        omega=frequencies, Hs=3, Tp=11, gamma=3.5, spectral_version="sim"
    )
    assert_almost_equal(result2, result_expected)

    # test the Hs on a higher resolution
    n_size = 10000
    frequencies = np.linspace(0, 10, n_size)
    delta_f = np.diff(frequencies)[0]
    hs_values = np.linspace(0, 10, num=21, endpoint=True)
    for hs_in in hs_values:
        # check version 0
        result = spectrum_jonswap(
            omega=frequencies, Hs=hs_in, Tp=11, gamma=3.5, spectral_version="sim"
        )
        hs_out = 4 * np.sqrt(result.sum() * delta_f)
        assert_almost_equal(hs_out, hs_in, decimal=2)

        # check version 1
        result = spectrum_jonswap(
            omega=frequencies, Hs=hs_in, Tp=11, gamma=3.5, spectral_version="sim"
        )
        hs_out = 4 * np.sqrt(result.sum() * delta_f)
        assert_almost_equal(hs_out, hs_in, decimal=2)

    hs_in = 10.0
    gamma_values = np.linspace(1, 6, num=21, endpoint=True)
    for gamma in gamma_values:
        # check version 0
        result = spectrum_jonswap(
            omega=frequencies, Hs=hs_in, Tp=11, gamma=gamma, spectral_version="dnv"
        )
        hs_out = 4 * np.sqrt(result.sum() * delta_f)
        assert_almost_equal(hs_out, hs_in, decimal=1)

        # check version 1
        result = spectrum_jonswap(
            omega=frequencies, Hs=hs_in, Tp=11, gamma=gamma, spectral_version="sim"
        )
        hs_out = 4 * np.sqrt(result.sum() * delta_f)
        assert_almost_equal(hs_out, hs_in, decimal=1)

    tp_values = np.linspace(4, 20, num=21, endpoint=True)
    for tp in tp_values:
        # check version 0
        result = spectrum_jonswap(
            omega=frequencies, Hs=hs_in, Tp=tp, gamma=3.3, spectral_version="dnv"
        )
        hs_out = 4 * np.sqrt(result.sum() * delta_f)
        assert_almost_equal(hs_out, hs_in, decimal=2)

        # check version 1
        result = spectrum_jonswap(
            omega=frequencies, Hs=hs_in, Tp=tp, gamma=3.3, spectral_version="sim"
        )
        hs_out = 4 * np.sqrt(result.sum() * delta_f)
        assert_almost_equal(hs_out, hs_in, decimal=2)


def test_omega_deep_water():
    n_size = 10
    wave_numbers = np.linspace(0, 2 * np.pi / 100, n_size)
    result = omega_deep_water(wave_number_vector=wave_numbers)
    result_expected = np.array(
        [
            0.0,
            0.261655,
            0.370036,
            0.4531997,
            0.52331,
            0.5850783,
            0.6409212,
            0.692274,
            0.7400721,
            0.784965,
        ]
    )
    assert_almost_equal(result, result_expected)


def test_spectrum_jonswap_k_domain():
    n_size = 10
    wave_numbers = np.linspace(0, 2 * np.pi / 100, n_size)
    result = spectrum_wave_k_domain(
        k_waves=wave_numbers,
        Hs=10,
        Tp=12,
        gamma=3.5,
        spectrum_type="jonswap",
        spectral_version="dnv",
    )
    result_expected = np.array(
        [
            0.0000000e00,
            4.5328519e-05,
            1.9150280e01,
            1.1190881e02,
            3.5913257e02,
            1.4081188e02,
            6.4523656e01,
            4.4606775e01,
            3.2823016e01,
            2.4618044e01,
        ]
    )
    assert_almost_equal(result, result_expected, decimal=5)

    result = spectrum_wave_k_domain(
        k_waves=wave_numbers,
        Hs=10,
        Tp=12,
        gamma=3.5,
        spectrum_type="jonswap",
        spectral_version="sim",
    )
    result_expected = np.array(
        [
            0.0000000e00,
            4.5226191e-05,
            1.9107049e01,
            1.1165618e02,
            3.5832184e02,
            1.4049400e02,
            6.4377996e01,
            4.4506077e01,
            3.2748920e01,
            2.4562470e01,
        ]
    )
    assert_almost_equal(result, result_expected, decimal=5)

    n_size = 10000
    wave_numbers = np.linspace(0, 10, n_size)
    delta_k = np.diff(wave_numbers)[0]
    hs_values = np.linspace(0, 10, num=21, endpoint=True)
    for hs_in in hs_values:
        # check version 0
        result = spectrum_wave_k_domain(
            k_waves=wave_numbers,
            Hs=hs_in,
            Tp=11,
            gamma=3.5,
            spectrum_type="jonswap",
            spectral_version="dnv",
        )
        hs_out = 4 * np.sqrt(result.sum() * delta_k)
        assert_almost_equal(hs_out, hs_in, decimal=2)

        # Spectrum_jonswap_k_domain_2 is the explicit verison of the transormed jonswap
        # spectrum. Should yield the same results as the
        result2 = spectrum_jonswap_k_domain_2(
            k_waves=wave_numbers, Hs=hs_in, Tp=11, gamma=3.5, spectral_version="dnv"
        )
        assert_almost_equal(result, result2)

        result = spectrum_wave_k_domain(
            k_waves=wave_numbers, Hs=hs_in, Tp=11, gamma=3.5, spectrum_type="gauss"
        )
        hs_out = 4 * np.sqrt(result.sum() * delta_k)
        assert_almost_equal(hs_out, hs_in, decimal=4)


def test_initialize_phase():
    n_size = 10
    wave_numbers = np.linspace(0, 2 * np.pi / 100, n_size)
    result = initialize_phase(k_waves=wave_numbers, seed=1)
    result_expected = np.array(
        [
            2.6202265e00,
            4.5259323e00,
            7.1863817e-04,
            1.8996116e00,
            9.2209446e-01,
            5.8018050e-01,
            1.1703074e00,
            2.1712221e00,
            2.4929636e00,
            3.3854854e00,
        ]
    )
    assert_almost_equal(result, result_expected)


def test_spectrum_to_complex_amplitudes():
    n_size = 10
    wave_numbers = np.linspace(0, 2 * np.pi / 100, n_size)
    spectrum = spectrum_wave_k_domain(
        k_waves=wave_numbers,
        Hs=10,
        Tp=12,
        gamma=3.5,
        spectrum_type="jonswap",
        spectral_version="dnv",
    )
    phase = initialize_phase(k_waves=wave_numbers, seed=1)

    result = spectrum_to_complex_amplitudes(
        wave_numbers, spectral_modulus=spectrum, phase=phase, mirror=False
    )

    result_expected = [
        -0.00000000e00 + 0.00000000e00j,
        -1.47478283e-04 - 7.81764461e-04j,
        5.17095937e-01 + 3.71604943e-04j,
        -4.03657860e-01 + 1.18304777e00j,
        1.35287488e00 + 1.78442321e00j,
        1.17273127e00 + 7.68639107e-01j,
        3.70050868e-01 + 8.74060956e-01j,
        -0.445890050764 + 0.651160605446j,
        -0.539490341563 + 0.408957133181j,
        -0.568936261313 - 0.141577804j,
    ]

    assert_almost_equal(result, result_expected)


def test_spectrum_to_complex_amplitudes_mirror():
    n_size = 5
    wave_numbers = np.linspace(0, 2 * np.pi / 100, n_size)
    spectrum = spectrum_wave_k_domain(
        k_waves=wave_numbers,
        Hs=10,
        Tp=12,
        gamma=3.5,
        spectrum_type="jonswap",
        spectral_version="dnv",
    )
    phase = initialize_phase(k_waves=wave_numbers, seed=1)

    result = spectrum_to_complex_amplitudes(
        wave_numbers, spectral_modulus=spectrum, phase=phase, mirror=True
    )

    result_expected = np.array(
        [
            -0.00000000 + 0.0j,
            -0.20406898 - 1.08174486j,
            2.83300855 + 0.00203591j,
            2.83300855 - 0.00203591j,
            -0.20406898 + 1.08174486j,
        ]
    )

    assert_almost_equal(result, result_expected)


def test_jonswap2D_complex_amplitudes():
    n_size = 4
    wave_numbers_x = np.linspace(0, 2 * np.pi / 100.0, n_size)
    wave_numbers_y = np.linspace(0, 2 * np.pi / 100.0, n_size)

    ak, os = spectrum2d_complex_amplitudes(
        kx_nodes=wave_numbers_x,
        ky_nodes=wave_numbers_y,
        Theta_0=pi / 6,
        Hs=2.1,
        Tp=12.2,
        gamma=3.3,
        Theta_s_spread_kx=5.0,
        mirror=True,
        seed=1,
    )

    ak_exp_real = np.array(
        [
            [0.0, -0.1108242, -0.0937341, -0.1108242],
            [-0.1718166, 0.2862183, -0.1266911, 0.0244118],
            [-0.0177563, 0.1111589, -0.1007618, 0.1111589],
            [-0.1718166, 0.0244118, -0.1266911, 0.2862183],
        ]
    )

    ak_exp_imag = np.array(
        [
            [0.0, 0.3248055, 0.1368857, -0.3248055],
            [0.0986876, 0.377518, 0.0960374, 0.0829592],
            [-0.0941239, 0.0728565, -0.0250742, -0.0728565],
            [-0.0986876, -0.0829592, -0.0960374, -0.377518],
        ]
    )

    os_exp = np.array([[1, 1, 1, -1], [1, 1, 1, 1], [1, 1, 1, -1], [-1, -1, -1, -1]])

    assert_almost_equal(ak.real, ak_exp_real)
    assert_almost_equal(ak.imag, ak_exp_imag)
    assert_almost_equal(os, os_exp)

    def check_symmetry(nx, ny, theta_zero, hs_target=2.1, debug_output=False):
        L = 1000
        dx = L / nx
        dy = L / ny
        kx_nodes = 2 * np.pi * np.fft.fftfreq(nx, dx)
        ky_nodes = 2 * np.pi * np.fft.fftfreq(ny, dy)
        ak, os = spectrum2d_complex_amplitudes(
            kx_nodes=kx_nodes,
            ky_nodes=ky_nodes,
            Theta_0=theta_zero,
            Hs=hs_target,
            Tp=12.2,
            gamma=3.3,
            Theta_s_spread_kx=5.0,
            mirror=True,
            seed=1,
        )

        # Parseval's theorem to calculate the variance from the complex fourier
        # coefficients divide by 2 as you only use half of the domain
        hs_estimate = 4 * np.sqrt(np.square(abs(ak)).sum()) / 2
        assert_almost_equal(hs_target, hs_estimate, decimal=1)

        if debug_output:
            print(f"calculating {nx} {ny} {theta_zero}")
            # print out hs. You can compare with amplitudes if you want
            # N = int(os.size / 2)
            # amplitudes = N * np.fft.ifft2(ak)
            # hs_et2 = 4 * np.std(amplitudes)
            print(
                "nx={} ny={} theta={}  {:.2f}  {:.2f}".format(
                    nx, ny, theta_zero, hs_target, hs_estimate
                )
            )

            # for debugging only if you want to view the maxtrix to see if it is
            # symmetric
            print()
            for j1 in range(ny):
                line = ""
                for i1 in range(nx):
                    line += "({:8.5f},{:8.5f}) ".format(
                        ak[i1][j1].real, ak[i1][j1].imag
                    )
                print(line)
        # This is an explicit loop over the complex amplitudes to check if it is point
        # mirrored
        for i1 in range(nx):
            for j1 in range(ny):
                if i1 == 0:
                    i2 = 0  # we are at the y-axis, which is mirrored to the - y-axis
                else:
                    i2 = nx - 1 - (i1 - 1)
                if j1 == 0:
                    j2 = 0  # we are at the x-axis, which is mirrored to the - x-axis
                else:
                    j2 = ny - 1 - (j1 - 1)
                ak_1 = ak[i1][j1]
                ak_2 = ak[i2][j2].conjugate()

                if i1 == i2 and j1 == j2:
                    # this is itself, skip as it is the same by definition and not a
                    # conjugate
                    continue

                # test if coefficient and its point mirrored partner are each other's
                # conjugate
                assert_almost_equal(ak_1, ak_2)

    n_size = 32
    # check both even and odd number of nodes in mesh and loop over a series of angles
    for i_off in [0, 1]:
        for j_off in [0, 1]:
            nx = n_size + i_off
            ny = 2 * n_size + j_off
            for theta_0 in np.linspace(0, 2 * pi, 7):
                check_symmetry(nx, ny, theta_zero=theta_0)


def test_omega_e_vs_omega():
    n_size = 10
    frequencies = np.linspace(0, 2.5, n_size)
    result = omega_e_vs_omega(frequencies, u_monitor=1.2)

    result_expected = np.array(
        [
            0.0,
            0.268336,
            0.5177883,
            0.748357,
            0.960042,
            1.1528435,
            1.3267613,
            1.4817954,
            1.6179459,
            1.7352128,
        ]
    )

    assert_almost_equal(result, result_expected)


def test_d_omega_e_prime():
    n_size = 10
    frequencies = np.linspace(0, 2.5, n_size)
    result = d_omega_e_prime(frequencies, velocity=2.2)
    result_expected = np.array(
        [
            1.0,
            0.875368,
            0.750736,
            0.6261041,
            0.5014721,
            0.3768401,
            0.2522081,
            0.1275761,
            0.0029441,
            -0.1216878,
        ]
    )
    assert_almost_equal(result, result_expected)


def test_omega_critical():
    n_size = 10
    om_c, index_c = omega_critical(velocity=2)
    om_c_exp = 2.4516625
    index_c_exp = 4
    assert_almost_equal([om_c], [om_c_exp])
    assert_equal([index_c], [None])

    frequencies = np.linspace(0, 5.5, n_size)
    om_c, index_c = omega_critical(velocity=2, omega=frequencies)
    assert_almost_equal([om_c], [om_c_exp])
    assert_equal([index_c], [index_c_exp])


def test_omega_vs_omega_e():
    result = omega_vs_omega_e(omega=1.2, velocity=1.2)

    result_expected = np.array([6.7109084, 1.4612999])

    assert_almost_equal(result, result_expected)


def test_spectrum_to_spectrum_encountered():
    n_size = 4
    frequencies = np.linspace(0, 2.5, n_size)
    spectrum_js = spectrum_jonswap(omega=frequencies, Hs=1, Tp=11, gamma=3.4)

    result, results2 = spectrum_to_spectrum_encountered(
        spectrum=spectrum_js, frequencies=frequencies, velocity=1
    )
    result_expected = np.array([0.0409422, 0.0212782, 0.0010317, 0.0004494])

    assert_almost_equal(result, result_expected)


def test_spectrum2d_to_spectrum2d_encountered():
    n_size = 4
    frequencies = np.linspace(0, 2.5, n_size)
    directions = np.linspace(0, 2 * np.pi, n_size, endpoint=False)
    spectrum_js = spectrum_jonswap(omega=frequencies, Hs=1, Tp=11, gamma=3.4)
    directional_dist = spreading_function(
        theta=directions, theta0=34, n_spreading_factor=2.3
    )
    ff_2d, dd_2d = np.meshgrid(frequencies, directions, indexing="ij")
    ss_om, dd_sp = np.meshgrid(spectrum_js, directional_dist, indexing="ij")
    s_2d = ss_om * dd_sp

    result = spectrum2d_to_spectrum2d_encountered(
        spectrum_2d=s_2d, frequencies=ff_2d, directions=dd_2d, velocity=0.1
    )
    result_expected = np.array(
        [
            [0.0000000e00, 6.3386893e-03, 1.8629551e-02, 0.0000000e00],
            [0.0000000e00, 3.2976248e-03, 9.7615486e-03, 0.0000000e00],
            [0.0000000e00, 1.4540844e-04, 5.0603899e-04, 0.0000000e00],
            [0.0000000e00, 3.4256523e-05, 1.1853199e-04, 0.0000000e00],
        ]
    )

    assert_almost_equal(result, result_expected)


def test_mask_out_of_range():
    n_size = 20
    wave_numbers = np.linspace(0, 2 * np.pi / 100, n_size)
    result = mask_out_of_range(kx=wave_numbers, kmin=0.01, kmax=0.03)

    result_expected = np.array(
        [
            False,
            False,
            False,
            False,
            True,
            True,
            True,
            True,
            True,
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
        ],
        dtype=bool,
    )

    assert_equal(result, result_expected)


def test_specspecs():
    n_size = 20
    frequencies = np.linspace(0, 2.5, n_size)
    spectrum_js = spectrum_jonswap(omega=frequencies, Hs=3, Tp=11, gamma=3.5)
    result = specspecs(frequency=frequencies, amplitude=spectrum_js)

    result_expected = (
        3,
        7,
        4,
        0.39473684210526316,
        0.92105263157894735,
        0.52631578947368418,
        1.6336106158889372,
        0.4975741649385097,
    )

    assert_almost_equal(result, result_expected)


def test_thetaspreadspecs():
    n_size = 20
    directions = np.linspace(0, 2 * np.pi, n_size, endpoint=False)
    dd_spread = spreading_function2(theta=directions, theta0=100)
    result = thetaspreadspecs(theta=directions, Dspread=dd_spread, areafraction=0.9)

    result_p1 = result[:7]
    result_p2 = result[7]

    result_expected_p1 = (
        15,
        1,
        4.7123889803846897,
        0.31415926535897931,
        5.6548667764616276,
        0.63910521149421706,
        0.99999999999999811,
    )

    result_expected_p2 = np.array(
        [
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            True,
            True,
            True,
            True,
        ],
        dtype=bool,
    )

    assert_almost_equal(result_p1, result_expected_p1)
    assert_equal(result_p2, result_expected_p2)


def test_rayleigh_pdf():
    n_size = 10
    frequencies = np.linspace(0, 2.5, n_size)
    result = rayleigh_pdf(omega=frequencies, sigma=3.1)
    result_expected = np.array(
        [
            0.0,
            0.007219,
            0.0143946,
            0.0214839,
            0.0284446,
            0.035236,
            0.041819,
            0.0481564,
            0.0542136,
            0.0599585,
        ]
    )
    assert_almost_equal(result, result_expected)


def test_rayleigh_cdf():
    n_size = 10
    frequencies = np.linspace(0, 2.5, n_size)
    result = rayleigh_cdf(omega=frequencies, sigma=3.1)
    result_expected = np.array(
        [
            1.0,
            0.9989969,
            0.9959935,
            0.9910078,
            0.9840699,
            0.975221,
            0.9645136,
            0.9520109,
            0.937786,
            0.9219212,
        ]
    )
    assert_almost_equal(result, result_expected)


def test_set_heading():
    n_size = 4
    np.random.seed(0)
    rao_data = np.random.rand(n_size, n_size)
    directions = np.linspace(0, 2 * np.pi, n_size, endpoint=False)
    result = set_heading(
        data=rao_data, directions=directions, heading=232, direction_axis=1
    )

    result_expected = np.array(
        [
            [0.5448832, 0.5488135, 0.7151894, 0.6027634],
            [0.891773, 0.4236548, 0.6458941, 0.4375872],
            [0.5288949, 0.9636628, 0.3834415, 0.791725],
            [0.0871293, 0.5680446, 0.9255966, 0.0710361],
        ]
    )
    assert_almost_equal(result, result_expected)


def main():
    """
    Run all the unit test here.

    This is only needed if the test_wave_spectra script is executed as a python script
    for debugging purpose only.
    In case that the setup test suit is run, this is not needed at all
    """
    test_omega_peak_jonswap()
    test_alpha_jonswap()
    test_spreading_function2()
    test_spreading_function()
    test_spectrum_gauss()
    test_spectrum_jonswap()
    test_omega_deep_water()
    test_spectrum_jonswap_k_domain()
    test_spectrum_to_complex_amplitudes()
    test_jonswap2D_complex_amplitudes()
    test_omega_e_vs_omega()
    test_d_omega_e_prime()
    test_omega_critical()
    test_omega_vs_omega_e()
    test_spectrum_to_spectrum_encountered()
    test_spectrum2d_to_spectrum2d_encountered()
    test_mask_out_of_range()
    test_specspecs()
    test_thetaspreadspecs()
    test_rayleigh_pdf()
    test_rayleigh_cdf()
    test_set_heading()


if __name__ == "__main__":
    main()



import matplotlib.pyplot as plt
from numpy import (sin, pi, linspace)
from numpy.testing import (assert_almost_equal)
from pymarine.utils.plotting import sub_plot_axis_to_2d


def test_sub_plot_axis_to_2d():
    def test_plot(nr, nc, x, y):
        # create a plot with nr x nc subplots
        fig, axis = plt.subplots(nrows=nr, ncols=nc)
        # turn the axis into a 2D list
        axis_2d = sub_plot_axis_to_2d(axis, n_rows=nr, n_cols=nc)
        # for all the plots, add a line and check if the plot line data equals the original data
        for i_row in range(nr):
            for j_col in range(nc):
                line, = axis_2d[i_row][j_col].plot(x, y)
                x_line, y_line = line.get_xydata().T
                assert_almost_equal(x_line, x)
                assert_almost_equal(y_line, y)

    x = linspace(0, 2 * pi, num=100)
    y = sin(x)
    # loop over a number of rows and columns ranging in between 1,2,3
    for nr in range(1, 4):
        for nc in range(1, 4):
            test_plot(nr, nc, x, y)



"""
Some numerical utilities used in other modules
"""

import re

import numpy as np
import scipy.io as spio


def ecdf2percentile(ecdf, percentile):
    """Calculate a percentile of an Empirical CDF function as returned by the
    `statsmodels` package

    Parameters
    ----------
    ecdf : :obj:`statsmodel.distributions.empirical_distribution.ECDF`
        Empirical CDF as a step function.
    percentile : float
        Percentile of the distribution

    Returns
    -------
    float:
       value where a fraction  of `percentile` will be lower

    Examples
    --------

    >>> import statsmodels.api as sm
    >>> np.random.seed(0)
    >>> number_of_observations = 100
    >>> # generate random data variing in between 0 and 100
    >>> x_data = 100 * np.random.rand(number_of_observations)
    >>> # calculate the cumulative distribution function of this random data
    >>> e_cdf = sm.distributions.empirical_distribution.ECDF(x_data)
    >>> # print some P values belonging to the percentiles
    >>> print("percentile {:2.1f} : ecdf ={:6.2f}".format(0.1,
    ... ecdf2percentile(ecdf=e_cdf, percentile=0.1)))
    percentile 0.1 : ecdf =  9.61
    >>> print("percentile {:2.1f} : ecdf ={:6.2f}".format(0.5,
    ... ecdf2percentile(ecdf=e_cdf, percentile=0.5)))
    percentile 0.5 : ecdf = 46.87
    >>> print("percentile {:2.1f} : ecdf ={:6.2f}".format(0.9,
    ... ecdf2percentile(ecdf=e_cdf, percentile=0.9)))
    percentile 0.9 : ecdf = 89.18
    """

    for i in range(ecdf.n):
        if ecdf.y[i] > percentile:
            return ecdf.x[i]


def get_parameter_list_key(parlist):
    """Utility for the qtgraphs Parameter Item

    Parameters
    ----------
    parlist :
        The parameter tree list (is an ordered list) contain all the values and current
        one

    Returns
    -------
    str
        The name belonging to the current value

    Notes
    -----
    The parameter tree widget has a field 'list' in which a list of values is given with
    corresponding integers. The current integer belonging to the parlist is obtained by
    parlist.value() however, to get the associated value of the key field is less
    straightforward In this routine it is retrieved
    """

    # the current value of the parlist
    value = parlist.value()

    # get the reverse list from the parlist
    reverselist = parlist.reverse

    # get the index belonging to the current value
    index = reverselist[0].index(value)

    # get the name of the key of this index
    keyname = reverselist[1][index]

    return keyname


def get_column_with_max_cumulative_value(data, regular_expression=".*"):
    """Find the column of a pandas DataFrame with the maximum cumulative value

    Parameters
    ----------
    data : DataFrame
        Data frame with the columns
    regular_expression : str
        Regular expression used to make a selection of columns to include. Default to
        '.*', which means that all columns are included

    Returns
    -------
    str or None:
        The name of the column with the maximum cumulative value or None if no columns
        were found

    Notes
    -----
    * Only the columns with a name obeying the regular expression are taken into account

    * An example of usage can be found in the fatigue monitoring software where we have
      data frames with damage over all the channels at a hot spots. If you want to
      obtained the channel with the maximum cumulative damage you can use this function

    Examples
    --------

    >>> import string
    >>> import pandas as pd
    >>> np.random.seed(0)
    >>> n_cols = 5
    >>> n_rows = 10

    Create a 10 x 5 data frame with random values with columns named as A, B, C, etc

    >>> data_frame = pd.DataFrame(np.random.random_sample((n_rows, n_cols)),
    ...                           columns=list(string.ascii_uppercase)[:n_cols])

    Obtain the name of the column with the maximum cumulative value

    >>> get_column_with_max_cumulative_value(data_frame)
    'D'

    Obtain the name of the column with the maximum cumulative value only including
    colums A, B and C

    >>> get_column_with_max_cumulative_value(data_frame, regular_expression="[ABC]")
    'C'

    """

    columns = list()
    for col in data.columns.values:
        if re.match(regular_expression, col):
            columns.append(col)
    if columns:
        name_of_maximum_column = data.fillna(0).sum()[columns[0] : columns[-1]].idxmax()
    else:
        name_of_maximum_column = None

    return name_of_maximum_column


def nans(shape, dtype=float):
    """Create an array filled with :class:`numpy.nan` values

    Parameters
    ----------
    shape : tuple
        Shape of the array of nans to create

    dtype : `numpy.dtype`, optional, default=float
        Type of the nan

    Returns
    -------
    ndarray
        Array filled with `numpy.nan` values

    Examples
    --------
    >>> nans((2, 3))
    array([[ nan,  nan,  nan],
           [ nan,  nan,  nan]])

    """
    # create an array of nans with give shape
    arr = np.empty(shape, dtype)
    arr.fill(np.nan)
    return arr


def get_nearest_index(data, value):
    """
    Find the index of the first occurrence of a value in a array with monotonically
    increasing values

    Parameters
    ----------
    data : array_like
        Array with monotonically increasing numbers
    value :
        Value of the point to give the index

    Returns
    -------
    int
        Index of floor cell where the values in the array first exceed the value

    Example
    -------
    >>> data_array = np.linspace(0, 10, 8, endpoint=False)
    >>> print(data_array)
    [ 0.    1.25  2.5   3.75  5.    6.25  7.5   8.75]
    >>> get_nearest_index(data_array, value=3)
    2
    >>> get_nearest_index(data_array, value=3.5)
    2
    >>> get_nearest_index(data_array, value=8.5)
    6
    >>> get_nearest_index(data_array, value=10)
    7
    >>> get_nearest_index(data_array, value=-1)
    0

    In case of a non-increasing array an AssertionError is raised

    >>> data_array = np.array([3, 0, 2, 4, 2.11, 1])
    >>> try:
    ...   get_nearest_index(data_array, value=2.1)
    ... except AssertionError as err:
    ...   print("An assertion error was raised")
    An assertion error was raised

    Raises
    ------
    AssertionError
        In case the array does not have monotonically increasing values

    Notes
    -----
    * Only arrays with monotonically increasing values are allowed.
    * In case the nearest index of an arbitrary array is needed, `find_inx_nearest_val`
      should be used
    * In case the `value` a larger than any value in the `data` array, the maximum index
      is returned, while in case the `value` is smaller than any value in the `data`
      array, a zero is return

    See Also
    --------
    find_idx_nearest_val: a function to get the index of the nearest value for an
    arbitraray array

    """
    data_array = np.asanyarray(data)
    if (np.diff(data_array) < 0).any():
        raise AssertionError(
            "Only monotonic increasing arrays are allowed. "
            "Use `find_inx_nearest_val` instead "
        )

    sgn = np.sign(np.asarray(data) - value)
    try:
        nearest_index = np.where(np.diff(sgn) > 0)[0][0]
    except IndexError:
        if sgn[0] > 0:
            nearest_index = 0
        else:
            nearest_index = data.size - 1

    return nearest_index


def find_idx_nearest_val(array, value):
    """Find the nearest index of a value in a array.

    Parameters
    ----------
    array : array_like
        an array with values
    value : float
        the value for which we want the nearest index

    Returns
    -------
    int
        The index of the nearest value to 'value' in array.
          the number of items in the array

    Examples
    --------
    >>> data_array = np.linspace(0, 10, 8, endpoint=False)
    >>> print(data_array)
    [ 0.    1.25  2.5   3.75  5.    6.25  7.5   8.75]
    >>> find_idx_nearest_val(data_array, value=3)
    2
    >>> find_idx_nearest_val(data_array, value=3.5)
    3
    >>> find_idx_nearest_val(data_array, value=8.5)
    7
    >>> find_idx_nearest_val(data_array, value=10)
    7
    >>> find_idx_nearest_val(data_array, value=-1)
    0

    In case of a non-increasing array with equal candidates, the last one is returned

    >>> data_array = np.array([3, 0, 2, 4, 2, 1])
    >>> find_idx_nearest_val(data_array, value=2.1)
    4
    >>> data_array = np.array([3, 0, 2, 4, 2, 1, 2])
    >>> find_idx_nearest_val(data_array, value=2.1)
    6

    If we change the first candidate, so it becomes the nearest, this one will be
    returned

    >>> data_array = np.array([3, 0, 2.09, 4, 2, 1])
    >>> find_idx_nearest_val(data_array, value=2.1)
    2

    Notes
    -----
    * In case that 2 or more items exist with the same distance from `value`, the
      *last* occurrence is returned
    * In case the value is outside the range of any value inside the array, either 0 or
      N-1 is return, with N the number of array elements of the input array

    See Also
    --------
    get_nearest_index:
        This function returns the first occurrence and only works for monotonically
        increasing arrays
    """
    idx_sorted = np.argsort(array)
    sorted_array = np.array(array[idx_sorted])
    idx = np.searchsorted(sorted_array, [value], side="left")
    try:
        # in case search sorted returns a list of values take the first one
        idx = idx[0]
    except IndexError:
        # it was not a list, so do't do anything
        pass

    if idx >= len(array):
        idx_nearest = idx_sorted[len(array) - 1]
    elif idx == 0:
        idx_nearest = idx_sorted[0]
    else:
        if abs(value - sorted_array[idx - 1]) < abs(value - sorted_array[idx]):
            idx_nearest = idx_sorted[idx - 1]
        else:
            idx_nearest = idx_sorted[idx]

    return idx_nearest


def get_range_from_string(range_string):
    """Analyse a range string to get the start, end and step and return a numpy array

    Parameters
    ----------
    range_string: str, start:end:[step]
        String representing the start, end and step size of a range

    Returns
    -------
    ndarray:
        1-D Array with points defined by range_string

    Examples
    --------
    >>> get_range_from_string("0:10:2")
    array([  0.,   2.,   4.,   6.,   8.,  10.])

    >>> get_range_from_string("0:7")
    array([ 0.,  1.,  2.,  3.,  4.,  5.,  6.,  7.])

    >>> get_range_from_string("3:4:0.2")
    array([ 3. ,  3.2,  3.4,  3.6,  3.8,  4. ])
    """
    values = range_string.split(":")
    try:
        start = values[0]
        stop = values[1]
        try:
            spacing = values[2]
        except IndexError:
            spacing = 1.0
    except IndexError:
        raise AssertionError("Usage supply string 'start:end:[step]'")

    try:
        start = float(start)
        stop = float(stop)
        spacing = float(spacing)
    except ValueError:
        raise AssertionError(
            "Could not convert one of the value to a float from string {}"
            "".format(range_string)
        )

    n_points = int((stop - start) / spacing) + 1

    return np.linspace(start, stop, n_points)


def make_2d_array_cyclic(data_2d, axis=0, add_constant=0.0):
    """Makes a 2D array period by copying the first row toward the end

    Parameters
    ----------
    data_2d : ndarray
        NxM array carrying the data. It is assumed that the data along one of the axis
        is periodic and that we want to make the array cyclic by copying the first row
        or column to the end
    axis: int
        Axis to make periodic. Default = 0
    add_constant : float
        Add this constant to the copy row or column to allow to add 2*pi or 360

    Returns
    -------
    ndarray
        Same 2d data array with one extra column (N+1 x M for axis == 0) or one extra
        row (N x M + 1 for axis == 1)


    Examples
    --------
    First make some 2D data array using mesh grid containing the direction we want to
    make periodic

    >>> directions = np.linspace(0, 360, 6, endpoint=False)
    >>> frequencies = np.linspace(0, 3, 3)
    >>> dd, ff = np.meshgrid(directions, frequencies)
    >>> dd
    array([[   0.,   60.,  120.,  180.,  240.,  300.],
           [   0.,   60.,  120.,  180.,  240.,  300.],
           [   0.,   60.,  120.,  180.,  240.,  300.]])

    The array `dd` is periodic along the axis = 0 direction. So copy the first column to
    the end

    >>> dd_periodic = make_2d_array_cyclic(dd)
    >>> dd_periodic
    array([[   0.,   60.,  120.,  180.,  240.,  300.,    0.],
           [   0.,   60.,  120.,  180.,  240.,  300.,    0.],
           [   0.,   60.,  120.,  180.,  240.,  300.,    0.]])

    We can do the same in case we have the transposed version, only we have to use the
    axis = 1 argument to pick the right axis

    >>> dd_tr = dd.T
    >>> dd_tr
    array([[   0.,    0.,    0.],
           [  60.,   60.,   60.],
           [ 120.,  120.,  120.],
           [ 180.,  180.,  180.],
           [ 240.,  240.,  240.],
           [ 300.,  300.,  300.]])
    >>> dd_tr_periodic = make_2d_array_cyclic(dd_tr, axis=1)
    >>> dd_tr_periodic
    array([[   0.,    0.,    0.],
           [  60.,   60.,   60.],
           [ 120.,  120.,  120.],
           [ 180.,  180.,  180.],
           [ 240.,  240.,  240.],
           [ 300.,  300.,  300.],
           [   0.,    0.,    0.]])

    For polar plotting it is required that the angle is increasing, which means that
    the last 0 needs to be 360. This can be established by using the `add_constant`
    option

    >>> dd_periodic = make_2d_array_cyclic(dd, add_constant=360.)
    >>> dd_periodic
    array([[   0.,   60.,  120.,  180.,  240.,  300.,  360.],
           [   0.,   60.,  120.,  180.,  240.,  300.,  360.],
           [   0.,   60.,  120.,  180.,  240.,  300.,  360.]])

    Notes
    -----
    This function can be used to ensure that a 2D data array with directions at one axis
    can be easily made periodic into the direction axis. This is a requirement to make a
    polar plot

    """

    if axis == 0:
        # Create a Mx1 array of the first column of the 2D data array
        first_col = data_2d[:, 0].reshape(data_2d.shape[0], 1) + add_constant
        data_2d_periodic = np.hstack((data_2d, first_col))
    elif axis == 1:
        # create a 1xN array of the first row of the 2D data array
        first_row = data_2d[0, :].reshape(1, data_2d.shape[1]) + add_constant
        data_2d_periodic = np.vstack((data_2d, first_row))
    else:
        raise AssertionError(f"Argument `axis` can only be 0 or 1. {axis} given.")

    return data_2d_periodic


def extrap1d(interpolator):
    """
    Extrapolate the interp1d function outside the boundaries

    .. deprecated:: 0.3.4

    Deprecated function, the scipy *interp1d* now can extrapolate as well. This function
    is maintained for backward compatibility.

    Parameters
    ----------
    interpolator: interp1d
        The interpolator object created with scipy

    Returns
    -------
    function
        New interpolator object that extrapolates values outside the range

    Examples
    --------

    Assumed you have a x and y array you want to interpolate. You can use the
    scipy interp1d function for that

    >>> from scipy.interpolate import interp1d

    >>> xp = np.linspace(0, 3, 4)
    >>> yp = xp**2
    >>> print(np.vstack((xp, yp)))
    [[ 0.  1.  2.  3.]
     [ 0.  1.  4.  9.]]

    Use the data samples to create a Interpolator *f_inter*

    >>> f_inter = interp1d(xp, yp)

    Interpolation on a new mesh within the boundaries of the previous mesh can be done
    as

    >>> xp_new = np.linspace(0, 3, 6)
    >>> yp_new = f_inter(xp_new)
    >>> print(np.vstack((xp_new, yp_new)))
    [[ 0.   0.6  1.2  1.8  2.4  3. ]
     [ 0.   0.6  1.6  3.4  6.   9. ]]

    However, perhaps you want to extend the boundaries outside the initial boundaries.
    In that case you can use extrap1 in order to extrapolate the function outside the
    boundaries. First create the Extrapolator using the scipy Interpolator

    >>> f_extra = extrap1d(f_inter)

    Now we can also extrapolate outside the range

    >>> xp_new2 = np.linspace(0, 4, 9)
    >>> yp_new2 = f_extra(xp_new2)
    >>> print(np.vstack((xp_new2, yp_new2)))
    [[  0.    0.5   1.    1.5   2.    2.5   3.    3.5   4. ]
     [  0.    0.5   1.    2.5   4.    6.5   9.   11.5  14. ]]

    In the latest scipy version extrapolation is possible with the *interp1d* function
    as well  :

    >>> f_extra2 = interp1d(xp, yp, fill_value="extrapolate")
    >>> yp_new3 = f_extra2(xp_new2)
    >>> print(np.vstack((yp_new2, yp_new3)))
    [[  0.    0.5   1.    2.5   4.    6.5   9.   11.5  14. ]
     [  0.    0.5   1.    2.5   4.    6.5   9.   11.5  14. ]]

    As you can see, the result is the same as the *extrap1d* function. The native
    scipy.interp1d with *extrapolate* as *fill_values* is recommended; the *extrap1d*
    function is only kept for backward compatibility and may be dropped soon

    We can plot the results and compare it with the original data line

    .. plot::

    >>> import matplotlib.pyplot as plt
    >>> l = plt.plot(xp, yp, "o", label="samples")
    >>> l = plt.plot(xp_new2, yp_new2, "-x", label="extrap1d")
    >>> xp_new3 = np.linspace(0, 4, 50)
    >>> yp_new3 = xp_new3**2
    >>> l = plt.plot(xp_new3, yp_new3, "-", label="original")
    >>> l = plt.legend()
    >>> plt.ion()
    >>> plt.show()

    Notes
    -----

    * In case you want to use interpolate with values outside of the boundaries, this
      function allows to extrapolate outside the boundaries as described here interp_

    References
    ----------

    http://stackoverflow.com/questions/2745329/
    how-to-make-scipy-interpolate-give-an-extrapolated-result-beyond-the-input-range
    https://stackoverflow.com/a/37172840/4515114

    .. _interp:
        http://stackoverflow.com/questions/2745329/
        how-to-make-scipy-interpolate-give-an-extrapolated-result-beyond-the-input-range
    """

    xs = interpolator.x
    ys = interpolator.y

    def pointwise(x):
        if x < xs[0]:
            return ys[0] + (x - xs[0]) * (ys[1] - ys[0]) / (xs[1] - xs[0])
        elif x > xs[-1]:
            return ys[-1] + (x - xs[-1]) * (ys[-1] - ys[-2]) / (xs[-1] - xs[-2])
        else:
            return interpolator(x)

    def ufunclike(xs):
        return np.array(list(map(pointwise, np.array(xs))))

    return ufunclike


def print_mat_nested(d, indent=0, nkeys=0):
    """
    Pretty print nested structures from .mat files

    Parameters
    ----------
    indent: int
        Number of spaces to indent
    nkeys: int
        Number of keys to show

    References
    ----------
    * http://stackoverflow.com/questions/3229419/
      pretty-printing-nested-dictionaries-in-python
    """

    # Subset dictionary to limit keys to print.  Only works on first level
    if nkeys > 0:
        d = {
            k: d[k] for k in sorted(d.keys())[:nkeys]
        }  # Dictionary comprehension: limit to first nkeys keys.

    if isinstance(d, dict):
        for key in sorted(d.keys()):  # loops through key, sort them
            value = d[key]
            print("{}{}{}".format("\t" * indent, "Key: ", str(key)))
            print_mat_nested(value, indent + 1)

    if (
        isinstance(d, np.ndarray) and d.dtype.names is not None
    ):  # Note: and short-circuits by default
        for n in d.dtype.names:  # This means it's a struct, it's bit of a kludge test.
            print("{}{}{}".format("\t" * indent, "Field: ", str(key)))
            print_mat_nested(d[n], indent + 1)


def loadmat(filename):
    """
    Load a matlab data file with a complex data structure

    Parameters
    ----------
    filename: str
        Name of the matlab file to import

    Returns
    -------
    dict:
        Dictionary with the complex matlab data structure

    Examples
    --------

    To read a matlab file do

    >>> import os
    >>> file_name = os.path.join("..", "data", "RAO_7.mat")
    >>> data = loadmat(filename=file_name)

    Now have a look at the contents

    >>> print_mat_nested(data)
    Key: DirRange
    Key: FreqRange
    Key: RAO
    Key: __globals__
    Key: __header__
    Key: __version__


    The data stored in the Dir Range field can be accessed as

    >>> data["DirRange"]
    array([  0,  15,  30,  45,  60,  75,  90, 105, 120, 135, 150, 165, 180,
           195, 210, 225, 240, 255, 270, 285, 300, 315, 330, 345], dtype=uint16)

    Notes
    -----

    * This function should be called instead of direct spio.loadmat as it cures the
      problem of not properly recovering python dictionaries from mat files. It calls
      the function check keys to cure all entries which are still mat-objects

    References
    ----------
    * http://pyhogs.github.io/reading-mat-files.html
    * http://stackoverflow.com/questions/7008608/
      scipy-io-loadmat-nested-structures-i-e-dictionaries
    """
    data = spio.loadmat(filename, struct_as_record=False, squeeze_me=True)
    return _check_keys(data)


def _check_keys(dict):
    """

    checks if entries in dictionary are mat-objects. If yes
    todict is called to change them to nested dictionaries

    Parameters
    ----------
    dict

    Returns
    -------
    dict
        Dictionary with the complex data structure

    """
    for key in dict:
        if isinstance(dict[key], spio.matlab.mio5_params.mat_struct):
            dict[key] = _todict(dict[key])
    return dict


def _todict(matobj):
    """
    Convert a matlab object to a dictionary

    Parameters
    ----------
    matobj:
        A matlab opject

    Returns
    -------
    dict
        Dictionary with a matlab object

    A recursive function which constructs from matobjects nested dictionaries
    """
    dict = {}
    for strg in matobj._fieldnames:
        elem = matobj.__dict__[strg]
        if isinstance(elem, spio.matlab.mio5_params.mat_struct):
            dict[strg] = _todict(elem)
        else:
            dict[strg] = elem
    return dict


if __name__ == "__main":
    import doctest

    doctest.testmod()
